package org.joget.sample;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import org.apache.http.HttpResponse;
import org.apache.http.NameValuePair;
import org.apache.http.ParseException;
import org.apache.http.client.entity.UrlEncodedFormEntity;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.client.methods.HttpRequestBase;
import org.apache.http.entity.StringEntity;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClients;
import org.apache.http.message.BasicNameValuePair;
import org.apache.http.util.EntityUtils;
import org.joget.apps.app.service.AppUtil;
import org.joget.apps.form.model.Element;
import org.joget.apps.form.model.FormAjaxOptionsBinder;
import org.joget.apps.form.model.FormBinder;
import org.joget.apps.form.model.FormData;
import org.joget.apps.form.model.FormLoadOptionsBinder;
import org.joget.apps.form.model.FormRow;
import org.joget.apps.form.model.FormRowSet;
import org.joget.apps.form.service.FormUtil;
import org.joget.commons.util.LogUtil;
import org.joget.commons.util.SecurityUtil;
import org.joget.commons.util.StringUtil;
import org.joget.plugin.property.service.PropertyUtil;
import org.joget.workflow.model.WorkflowAssignment;
import org.joget.workflow.util.WorkflowUtil;
import org.json.JSONException;
import org.json.JSONObject;

public class JsonCallFormOptionsBinder extends FormBinder implements FormLoadOptionsBinder, FormAjaxOptionsBinder {

    private final static String MESSAGE_PATH = "messages/JsonCallFormOptionsBinder";
    
    public String getName() {
        return "JSON Call Form Options Binder";
    }

    public String getVersion() {
        return "6.0.0";
    }

    public String getDescription() {
        return "Reads a JSON feed URL to get form options";
    }

    public String getLabel() {
        return getName();
    }

    public String getClassName() {
        return getClass().getName();
    }

    public String getPropertyOptions() {
        // Use "%s" to annotate in json where Ajax cascade option should be placed at
        String useAjax = "";
        if (SecurityUtil.getDataEncryption() != null && SecurityUtil.getNonceGenerator() != null) {
            useAjax = ",{name:'useAjax',label:'@@form.defaultformoptionbinder.useAjax@@',type:'checkbox',value :'false',options :[{value :'true',label :''}]}";
        }
        Object[] arguments = new Object[]{useAjax};
        String json = AppUtil.readPluginResource(getClass().getName(), "/properties/JsonCallFormOptionsBinder.json", arguments, true, MESSAGE_PATH);
        return json;
    }
    
    public FormRowSet load(Element element, String primaryKey, FormData formData) {
        return loadAjaxOptions(null);
    }

    public boolean useAjax() {
        return "true".equalsIgnoreCase(getPropertyString("useAjax"));
    }
    
    //If user design options to be dependent to another field, dependencyValues holds a String Array of the current value of referenced form element
    public FormRowSet loadAjaxOptions(String[] dependencyValues) {
        FormRowSet results = new FormRowSet();
        results.setMultiRow(true);
        String baseObjName = getPropertyString("baseObjName");
        String valueObjName = getPropertyString("valueObjName");
        String labelObjName = getPropertyString("labelObjName");
        String groupObjName = getPropertyString("groupObjName");
        Map jsonObject = new HashMap();
        
        try {
            //Attempt to get JSON data, if FAIL, return blank options
            if (useAjax() && dependencyValues != null) {
                for (String s : dependencyValues) {
                    if (s != null && !s.isEmpty()) {
                        Map tempMap = getJsonData(s);
                        if (tempMap != null && tempMap.size() > 0) {
                            jsonObject.putAll(getJsonData(s));
                        } else {
                            LogUtil.info(getClass().getName(), "Check your connection towards the JSON server.");
                        }
                    }
                }
            } else if (useAjax() && dependencyValues == null) {
                LogUtil.info(getClass().getName(), "Remember to use #dv# and assign dependency field ID, to use AJAX cascade options.");
            }
                
            if (!useAjax()) {
                jsonObject = getJsonData("");
            }
            
            if (jsonObject != null) {
                //Add empty option if configured true
                if ("true".equalsIgnoreCase(getPropertyString("addEmptyOption"))) {
                    FormRow row = new FormRow();
                    row.setProperty(FormUtil.PROPERTY_VALUE, "");
                    row.setProperty(FormUtil.PROPERTY_LABEL, getPropertyString("emptyLabel"));
                    row.setProperty(FormUtil.PROPERTY_GROUPING, "");
                    results.add(row);
                }

                //Extract intended results from valid JSON
                if (baseObjName != null && !baseObjName.isEmpty()) {
                    if (getObjectFromMap(baseObjName, jsonObject) instanceof HashMap) {
                        FormRow row = new FormRow();
                        String value = (String) getObjectFromMap(valueObjName.replace(baseObjName, baseObjName + "[" + 0 + "]"), jsonObject);
                                String label = (String) getObjectFromMap(labelObjName.replace(baseObjName, baseObjName + "[" + 0 + "]"), jsonObject);
                                String grouping = "";
                                if (groupObjName != null && !groupObjName.isEmpty()) {
                                    grouping = (String) getObjectFromMap(groupObjName.replace(baseObjName, baseObjName + "[" + 0 + "]"), jsonObject);
                                }
                                if (value != null) {
                                    row.setProperty(FormUtil.PROPERTY_VALUE, value);
                                } else {
                                    LogUtil.info(getClass().getName(), "Invalid Value Object Name.");
                                }
                                if (label != null) {
                                    row.setProperty(FormUtil.PROPERTY_LABEL, label);
                                } else {
                                    LogUtil.info(getClass().getName(), "Invalid Label Object Name.");
                                }
                                if (grouping != null) {
                                    row.setProperty(FormUtil.PROPERTY_GROUPING, grouping);
                                } else {
                                    LogUtil.info(getClass().getName(), "Invalid Grouping Object Name.");
                                }
                                results.add(row);

                    } else {
                        Object[] baseObjectArray = (Object[]) getObjectFromMap(baseObjName, jsonObject);
                        if (baseObjectArray != null && baseObjectArray.length > 0) {
                            for (int i = 0; i < baseObjectArray.length; i++) {
                                FormRow row = new FormRow();
                                String value = (String) getObjectFromMap(valueObjName.replace(baseObjName, baseObjName + "[" + i + "]"), jsonObject);
                                String label = (String) getObjectFromMap(labelObjName.replace(baseObjName, baseObjName + "[" + i + "]"), jsonObject);
                                String grouping = "";
                                if (groupObjName != null && !groupObjName.isEmpty()) {
                                    grouping = (String) getObjectFromMap(groupObjName.replace(baseObjName, baseObjName + "[" + i + "]"), jsonObject);
                                }
                                if (value != null) {
                                    row.setProperty(FormUtil.PROPERTY_VALUE, value);
                                } else {
                                    LogUtil.info(getClass().getName(), "Invalid Value Object Name.");
                                }
                                if (label != null) {
                                    row.setProperty(FormUtil.PROPERTY_LABEL, label);
                                } else {
                                    LogUtil.info(getClass().getName(), "Invalid Label Object Name.");
                                }
                                if (grouping != null) {
                                    row.setProperty(FormUtil.PROPERTY_GROUPING, grouping);
                                } else {
                                    LogUtil.info(getClass().getName(), "Invalid Grouping Object Name.");
                                }
                                if (value == null || label == null || grouping == null) {
                                    break;
                                }
                                results.add(row);
                            }
                        }
                    }
                } else {
                        //Returns prematurely when response is just slow
                        LogUtil.info(getClass().getName(), "Invalid Base Object Name.");
                }
            }
        } catch (Exception e) {
            LogUtil.error(getClass().getName(), e, "");
            FormRowSet dummyBlankSet = new FormRowSet();
            return dummyBlankSet;
        }
        return results;
    }
    
    public Map getJsonData(String dependencyValue) {
        WorkflowAssignment wfAssignment = (WorkflowAssignment) getProperty("workflowAssignment");

        String jsonUrl = (String) getPropertyString("jsonUrl");
        CloseableHttpClient client = null;
        HttpRequestBase request = null;
        Map jsonMap = null;
        
        //Replace #dv# with dependency value for URL if present
        jsonUrl = checkAndAppendDvHash(jsonUrl, dependencyValue);
        
        try {
            client = HttpClients.createDefault();

            jsonUrl = WorkflowUtil.processVariable(jsonUrl, "", wfAssignment);

            jsonUrl = StringUtil.encodeUrlParam(jsonUrl);

            if ("true".equalsIgnoreCase(getPropertyString("debugMode"))) {
                LogUtil.info(getClass().getName(), ("post".equalsIgnoreCase(getPropertyString("requestType"))?"POST":"GET") + " : " + jsonUrl);
            }
            
            if ("post".equalsIgnoreCase(getPropertyString("requestType"))) {
                request = new HttpPost(jsonUrl);
                
                if ("jsonPayload".equals(getPropertyString("postMethod"))) {
                    JSONObject obj = new JSONObject();
                    Object[] paramsValues = (Object[]) getProperty("params");
                    for (Object o : paramsValues) {
                        Map mapping = (HashMap) o;
                        String name  = mapping.get("name").toString();
                        name = checkAndAppendDvHash(name, dependencyValue);
                        String value = mapping.get("value").toString();
                        value = checkAndAppendDvHash(value, dependencyValue);
                        obj.accumulate(name, WorkflowUtil.processVariable(value, "", wfAssignment));
                    }

                    StringEntity requestEntity = new StringEntity(obj.toString(4));
                    ((HttpPost) request).setEntity(requestEntity);
                    request.setHeader("Content-type", "application/json");
                    if ("true".equalsIgnoreCase(getPropertyString("debugMode"))) {
                        LogUtil.info(getClass().getName(), "JSON Payload : " + obj.toString(4));
                    }
                } else if ("custom".equals(getPropertyString("postMethod"))) {
                    String customPayload = checkAndAppendDvHash(getPropertyString("customPayload"), dependencyValue);
                    StringEntity requestEntity = new StringEntity(customPayload);
                    ((HttpPost) request).setEntity(requestEntity);
                    request.setHeader("Content-type", "application/json");
                    if ("true".equalsIgnoreCase(getPropertyString("debugMode"))) {
                        LogUtil.info(getClass().getName(), "Custom JSON Payload : " + customPayload);
                    }
                } else {
                    List<NameValuePair> urlParameters = new ArrayList<NameValuePair>();
                    Object[] paramsValues = (Object[]) getProperty("params");
                    for (Object o : paramsValues) {
                        Map mapping = (HashMap) o;
                        String name  = mapping.get("name").toString();
                        name = checkAndAppendDvHash(name, dependencyValue);
                        String value = mapping.get("value").toString();
                        value = checkAndAppendDvHash(value, dependencyValue);
                        urlParameters.add(new BasicNameValuePair(name, WorkflowUtil.processVariable(value, "", wfAssignment)));
                        if ("true".equalsIgnoreCase(getPropertyString("debugMode"))) {
                            LogUtil.info(getClass().getName(), "Adding param " + name + " : " + value);
                        }
                    }
                    ((HttpPost) request).setEntity(new UrlEncodedFormEntity(urlParameters));
                }
            } else {
                request = new HttpGet(jsonUrl);
            }
            
            Object[] paramsValues = (Object[]) getProperty("headers");
            for (Object o : paramsValues) {
                Map mapping = (HashMap) o;
                String name  = mapping.get("name").toString();
                name = checkAndAppendDvHash(name, dependencyValue);
                String value = mapping.get("value").toString();
                value = checkAndAppendDvHash(value, dependencyValue);
                if (name != null && !name.isEmpty() && value != null && !value.isEmpty()) {
                    request.setHeader(name, value);
                    if ("true".equalsIgnoreCase(getPropertyString("debugMode"))) {
                        LogUtil.info(getClass().getName(), "Adding request header " + name + " : " + value);
                    }
                }
            }
            
            //Execute JSON REST call request
            HttpResponse response = client.execute(request);
            
            if ("true".equalsIgnoreCase(getPropertyString("debugMode"))) {
                LogUtil.info(getClass().getName(), jsonUrl + " returned with status : " + response.getStatusLine().getStatusCode());
            }

            String jsonResponse = EntityUtils.toString(response.getEntity(), "UTF-8");
            
            //Append "response" as base object if JSON does not have any base
            if (jsonResponse != null && !jsonResponse.isEmpty()) {
                if (jsonResponse.startsWith("[") && jsonResponse.endsWith("]")) {
                    jsonResponse = "{ \"response\" : " + jsonResponse + " }";
                }
                if ("true".equalsIgnoreCase(getPropertyString("debugMode"))) {
                    LogUtil.info(getClass().getName(), jsonResponse);
                }
                //Extract JSON object as a map
                jsonMap = PropertyUtil.getProperties(new JSONObject(jsonResponse));
            }            
        } catch (IOException ex) {
            LogUtil.error(getClass().getName(), ex, "Unable to execute RESTful call request. Check your connection or URL.");
        } catch (ParseException ex) {
            LogUtil.error(getClass().getName(), ex, "Unable to parse JSON to String for further processing.");
        } catch (JSONException ex) {
            LogUtil.error(getClass().getName(), ex, "Invalid JSON. Check JSON URL.");
        } finally {
            try {
                if (request != null) {
                    request.releaseConnection();
                }
                if (client != null) {
                    client.close();
                }
            } catch (IOException ex) {
                LogUtil.error(getClass().getName(), ex, "Unable to terminate RESTful call connection or connection does not exist.");
            }
        }
        return jsonMap;
    }
    
    //Check and replaces any #dv# occurences in string
    private String checkAndAppendDvHash(String string, String dv) {
            return string.replaceAll("#dv#", dv);
    }
    
    protected Object getObjectFromMap(String key, Map object) {
        try {
            if (key.contains(".")) {
                String subKey = key.substring(key.indexOf(".") + 1);
                key = key.substring(0, key.indexOf("."));

                Map tempObject = (Map) getObjectFromMap(key, object);

                if (tempObject != null) {
                    return getObjectFromMap(subKey, tempObject);
                }
            } else {
                if (key.contains("[") && key.contains("]")) {
                    String tempKey = key.substring(0, key.indexOf("["));
                    int number = Integer.parseInt(key.substring(key.indexOf("[") + 1, key.indexOf("]")));
                    if (object.get(tempKey) instanceof HashMap) {
                        Map tempObjectArray = (Map) object.get(tempKey);
                        if (tempObjectArray != null) {
                            return tempObjectArray;
                        }
                    } else {
                        Object tempObjectArray[] = (Object[]) object.get(tempKey);
                        if (tempObjectArray != null && tempObjectArray.length > number) {
                            return tempObjectArray[number];
                        }
                    }
                } else {
                    return object.get(key);
                }
            }
        } catch (Exception e) {
            LogUtil.error(getClass().getName(), e, "");
        }
        return null;
    }
}